@@ -121,6 +121,7 @@ void *_gpgrt_calloc (size_t n, size_t m);
 char *_gpgrt_strdup (const char *string);
 char *_gpgrt_strconcat (const char *s1, ...) GPGRT_ATTR_SENTINEL(0);
 void _gpgrt_free (void *a);
+void _gpgrt_wipememory (void *ptr, size_t len);
 /* The next is only to be used by visibility.c.  */
 char *_gpgrt_strconcat_core (const char *s1, va_list arg_ptr);
 
@@ -234,6 +235,7 @@ typedef enum
   {
     BACKEND_MEM,
     BACKEND_FD,
+    BACKEND_SOCK,
     BACKEND_W32,
     BACKEND_FP,
     BACKEND_USER,
@@ -301,6 +303,7 @@ struct _gpgrt_stream_internal
   unsigned int stdstream_fd:2;   /* 0, 1 or 2 for a standard stream.  */
   unsigned int printable_fname_inuse: 1;  /* es_fname_get has been used.  */
   unsigned int samethread: 1;    /* The "samethread" mode keyword.  */
+  unsigned int wipe: 1;          /* The "wipe" mode keyword.  */
   size_t print_ntotal;           /* Bytes written from in print_writer. */
   notify_list_t onclose;         /* On close notify function list.  */
 };
@@ -534,6 +537,7 @@ void _gpgrt_log_set_sink (const char *name, estream_t stream, int fd);
 void _gpgrt_log_set_socket_dir_cb (const char *(*fnc)(void));
 void _gpgrt_log_set_pid_suffix_cb (int (*cb)(unsigned long *r_value));
 void _gpgrt_log_set_prefix (const char *text, unsigned int flags);
+void _gpgrt_add_post_log_func (void (*f)(int));
 const char *_gpgrt_log_get_prefix (unsigned int *flags);
 int  _gpgrt_log_test_fd (int fd);
 int  _gpgrt_log_get_fd (void);
@@ -561,7 +565,7 @@ void _gpgrt_log_flush (void);
 void _gpgrt_logv_printhex (const void *buffer, size_t length,
                            const char *fmt, va_list arg_ptr);
 void _gpgrt_log_printhex (const void *buffer, size_t length,
-                          const char *fmt, ...) GPGRT_ATTR_PRINTF(3,4);;
+                          const char *fmt, ...) GPGRT_ATTR_PRINTF(3,4);
 
 void _gpgrt_logv_clock (const char *fmt, va_list arg_ptr);
 void _gpgrt_log_clock (const char *fmt, ...) GPGRT_ATTR_PRINTF(1,2);
@@ -626,127 +630,115 @@ gpg_err_code_t _gpgrt_make_pipe (int filedes[2], estream_t *r_fp,
                                  int direction, int nonblock);
 
 /* Convenience macros to create a pipe.  */
-#define _gpgrt_create_pipe(a)              _gpgrt_make_pipe ((a),NULL, 0,    0);
-#define _gpgrt_create_inbound_pipe(a,b,c)  _gpgrt_make_pipe ((a), (b), -1, (c));
-#define _gpgrt_create_outbound_pipe(a,b,c) _gpgrt_make_pipe ((a), (b),  1, (c));
+#define _gpgrt_create_pipe(a)              _gpgrt_make_pipe ((a),NULL, 0,    0)
+#define _gpgrt_create_inbound_pipe(a,b,c)  _gpgrt_make_pipe ((a), (b), -1, (c))
+#define _gpgrt_create_outbound_pipe(a,b,c) _gpgrt_make_pipe ((a), (b),  1, (c))
 
+/* Actions (at spawning a child process), which is OS-specific.  */
+gpg_err_code_t _gpgrt_spawn_actions_new (gpgrt_spawn_actions_t *r_act);
+void _gpgrt_spawn_actions_release (gpgrt_spawn_actions_t act);
+void _gpgrt_spawn_actions_set_env_rev (gpgrt_spawn_actions_t,
+                                       const char *const*);
+#ifdef HAVE_W32_SYSTEM
+void _gpgrt_spawn_actions_set_envvars (gpgrt_spawn_actions_t, char *);
+void _gpgrt_spawn_actions_set_redirect (gpgrt_spawn_actions_t,
+                                        void *, void *, void *);
+void _gpgrt_spawn_actions_set_inherit_handles (gpgrt_spawn_actions_t, void **);
+#else
+void _gpgrt_spawn_actions_set_environ (gpgrt_spawn_actions_t, char **);
+void _gpgrt_spawn_actions_set_redirect (gpgrt_spawn_actions_t, int, int, int);
+void _gpgrt_spawn_actions_set_inherit_fds (gpgrt_spawn_actions_t,
+                                           const int *);
+void _gpgrt_spawn_actions_set_atfork (gpgrt_spawn_actions_t, void (*)(void *),
+                                      void *);
+#endif
 
-/* Fork and exec the program PGMNAME.
- *
- * If R_INFP is NULL connect stdin of the new process to /dev/null; if
- * it is not NULL store the address of a pointer to a new estream
- * there. If R_OUTFP is NULL connect stdout of the new process to
- * /dev/null; if it is not NULL store the address of a pointer to a
- * new estream there.  If R_ERRFP is NULL connect stderr of the new
- * process to /dev/null; if it is not NULL store the address of a
- * pointer to a new estream there.  On success the pid of the new
- * process is stored at PID.  On error -1 is stored at PID and if
- * R_OUTFP or R_ERRFP are not NULL, NULL is stored there.
+/* Spawn the program PGMNAME with ARGV1.
  *
  * The arguments for the process are expected in the NULL terminated
- * array ARGV.  The program name itself should not be included there.
- * If PREEXEC is not NULL, the given function will be called right
- * before the exec.
+ * array ARGV1.  The program name itself should not be included there.
  *
- * IF EXCEPT is not NULL, it is expected to be an ordered list of file
- * descriptors, terminated by an entry with the value (-1).  These
- * file descriptors won't be closed before spawning a new program.
- *
- * Returns 0 on success or an error code.  Calling gpgrt_wait_process
- * and gpgrt_release_process is required if the function succeeded.
+ * Returns 0 on success or an error code.
  *
  * FLAGS is a bit vector:
  *
- * GPGRT_SPAWN_NONBLOCK
- *        If set the two output streams are created in non-blocking
- *        mode and the input stream is switched to non-blocking mode.
- *        This is merely a convenience feature because the caller
- *        could do the same with gpgrt_set_nonblock.  Does not yet
- *        work for Windows.
- *
- * GPGRT_SPAWN_DETACHED
+ * GPGRT_PROCESS_DETACHED
  *        If set the process will be started as a background process.
- *        This flag is only useful under W32 (but not W32CE) systems,
- *        so that no new console is created and pops up a console
- *        window when starting the server.  Does not work on W32CE.
+ *        No furthrer interaction between caller is expected.
+ *
+ * GPGRT_PROCESS_STDIN_PIPE
+ * GPGRT_PROCESS_STDOUT_PIPE
+ * GPGRT_PROCESS_STDERR_PIPE
+ *        Spawned child process is connected by pipes to the parent.
+ *
+ * GPGRT_PROCESS_STDIN_KEEP
+ * GPGRT_PROCESS_STDOUT_KEEP
+ * GPGRT_PROCESS_STDERR_KEEP
+ *        Do not assign /dev/null to a non-required standard file
+ *        descriptor.
+ *
+ * If ACT is not null, it specifies OS-specific internal actions at
+ * spawning a process; User may specify setting of environment
+ * variables, redirect of standard io, and inheriting file
+ * descriptors.
+ *
+ * If R_PROCESS is not null, the process object is returned into it.
+ * In this case, calling gpgrt_process_process is required if the
+ * function succeeded.
+ *
+ * If R_PROCESS is null, it waits for the child process's termination.
  *
- * GPGRT_SPAWN_RUN_ASFW
- *        On W32 (but not on W32CE) run AllowSetForegroundWindow for
- *        the child.  Note that due to unknown problems this actually
- *        allows SetForegroundWindow for all children of this process.
  */
-gpg_err_code_t
-_gpgrt_spawn_process (const char *pgmname, const char *argv[],
-                      int *execpt, void (*preexec)(void), unsigned int flags,
-                      estream_t *r_infp,
-                      estream_t *r_outfp,
-                      estream_t *r_errfp,
-                      pid_t *pid);
+gpg_err_code_t _gpgrt_process_spawn (const char *pgmname, const char *argv1[],
+                                     unsigned int flags,
+                                     gpgrt_spawn_actions_t act,
+                                     gpgrt_process_t *r_process);
 
+/* Kill a process; that is send an appropriate signal to the PROCESS.  */
+gpg_err_code_t _gpgrt_process_terminate (gpgrt_process_t process);
 
-/* Simplified version of gpgrt_spawn_process.  This function forks and
- * then execs PGMNAME, while connecting INFD to stdin, OUTFD to stdout
- * and ERRFD to stderr (any of them may be -1 to connect them to
- * /dev/null).  The arguments for the process are expected in the NULL
- * terminated array ARGV.  The program name itself should not be
- * included there.  Calling gpgrt_wait_process and
- * gpgrt_release_process is required.  Returns 0 on success or an
- * error code. */
-gpg_err_code_t _gpgrt_spawn_process_fd (const char *pgmname,
-                                        const char *argv[],
-                                        int infd, int outfd, int errfd,
-                                        pid_t *pid);
+gpg_err_code_t _gpgrt_process_get_fds (gpgrt_process_t process,
+                                       unsigned int flags,
+                                       int *r_fd_in, int *r_fd_out,
+                                       int *r_fd_err);
 
-/* Spawn a new process and immediately detach from it.  The name of
- * the program to exec is PGMNAME and its arguments are in ARGV (the
- * programname is automatically passed as first argument).
- * Environment strings in ENVP are set.  An error is returned if
- * pgmname is not executable; to make this work it is necessary to
- * provide an absolute file name.  */
-gpg_err_code_t _gpgrt_spawn_process_detached (const char *pgmname,
-                                              const char *argv[],
-                                              const char *envp[] );
+/*
+ * FLAGS can be:
+ *
+ * GPGRT_PROCESS_STREAM_NONBLOCK
+ *        If set the two stream(s) are created in non-blocking mode.
+ */
+gpg_err_code_t _gpgrt_process_get_streams (gpgrt_process_t process,
+                                           unsigned int flags,
+                                           gpgrt_stream_t *r_fp_in,
+                                           gpgrt_stream_t *r_fp_out,
+                                           gpgrt_stream_t *r_fp_err);
 
-/* If HANG is true, waits for the process identified by PID to exit;
- * if HANG is false, checks whether the process has terminated.
- * PGMNAME should be the same as supplied to the spawn function and is
- * only used for diagnostics.  Return values:
+gpg_err_code_t _gpgrt_process_vctl (gpgrt_process_t process,
+                                    unsigned int request, va_list arg_ptr);
+
+/* If HANG is true, waits for the process identified by PROCESS to
+ * exit; if HANG is false, checks whether the process has terminated.
+ * Return values:
  *
  * 0
- *     The process exited successful.  0 is stored at R_EXITCODE.
- *
- * GPG_ERR_GENERAL
- *     The process exited without success.  The exit code of process
- *     is then stored at R_EXITCODE.  An exit code of -1 indicates
- *     that the process terminated abnormally (e.g. due to a signal).
+ *     The process exited.
  *
  * GPG_ERR_TIMEOUT
  *     The process is still running (returned only if HANG is false).
  *
- * GPG_ERR_INV_VALUE
- *     An invalid PID has been specified.
- *
- * Other error codes may be returned as well.  Unless otherwise noted,
- * -1 will be stored at R_EXITCODE.  R_EXITCODE may be passed as NULL
- * if the exit code is not required (in that case an error message will
- * be printed).  Note that under Windows PID is not the process id but
- * the handle of the process.  */
-gpg_err_code_t _gpgrt_wait_process (const char *pgmname, pid_t pid, int hang,
-                                    int *r_exitcode);
+ * GPG_ERR_ECHILD / GPG_ERR_GENERAL
+ *     Waiting the process failed.
+ */
+gpg_err_code_t _gpgrt_process_wait (gpgrt_process_t process, int hang);
 
-/* Like _gpgrt_wait_process, but for COUNT processes.  */
-gpg_err_code_t _gpgrt_wait_processes (const char **pgmnames, pid_t *pids,
-                                      size_t count, int hang, int *r_exitcodes);
+/* Release the process identified by PROCESS.  */
+void _gpgrt_process_release (gpgrt_process_t process);
 
-/* Kill a process; that is send an appropriate signal to the process.
- * gpgrt_wait_process must be called to actually remove the process
- * from the system.  An invalid PID is ignored.  */
-void _gpgrt_kill_process (pid_t pid);
 
-/* Release the process identified by PID.  This function is actually
- * only required for Windows but it does not harm to always call it.
- * It is a nop if PID is invalid.  */
-void _gpgrt_release_process (pid_t pid);
+/* Close all file resources (descriptors), except KEEP_FDS.  */
+void _gpgrt_close_all_fds (int from, const int *keep_fds,
+                           int fallback_max_fds);
 
 
 /*
@@ -769,9 +761,53 @@ void _gpgrt_set_confdir (int what, const char *name);
 int _gpgrt_cmp_version (const char *a, const char *b, int level);
 
 
+/*
+ * strlist.c
+ */
+void _gpgrt_strlist_free (gpgrt_strlist_t sl);
+gpgrt_strlist_t _gpgrt_strlist_add (gpgrt_strlist_t *list,
+                                    const char *string, unsigned int flags);
+gpgrt_strlist_t _gpgrt_strlist_tokenize (gpgrt_strlist_t *list,
+                                         const char *string, const char *delim,
+                                         unsigned int flags);
+gpgrt_strlist_t _gpgrt_strlist_copy (gpgrt_strlist_t list);
+gpgrt_strlist_t _gpgrt_strlist_rev (gpgrt_strlist_t *list);
+gpgrt_strlist_t _gpgrt_strlist_prev (gpgrt_strlist_t head,
+                                     gpgrt_strlist_t node);
+gpgrt_strlist_t _gpgrt_strlist_last (gpgrt_strlist_t node);
+char *_gpgrt_strlist_pop (gpgrt_strlist_t *list);
+gpgrt_strlist_t _gpgrt_strlist_find (gpgrt_strlist_t haystack,
+                                     const char *needle);
+
+/*
+ * name-value.c
+ */
+gpgrt_nvc_t _gpgrt_nvc_new (unsigned int flags);
+void _gpgrt_nvc_release (gpgrt_nvc_t cont);
+int _gpgrt_nvc_get_flag (gpgrt_nvc_t cont, unsigned int flags, int clear);
+gpg_err_code_t _gpgrt_nvc_add (gpgrt_nvc_t cont,
+                               const char *name, const char *value);
+gpg_err_code_t _gpgrt_nvc_set (gpgrt_nvc_t cont,
+                               const char *name, const char *value);
+gpg_err_code_t _gpgrt_nve_set (gpgrt_nvc_t cont, gpgrt_nve_t e,
+                               const char *value);
+void _gpgrt_nvc_delete (gpgrt_nvc_t cont, gpgrt_nve_t entry, const char *name);
+gpgrt_nve_t _gpgrt_nvc_lookup (gpgrt_nvc_t cont, const char *name);
+gpg_err_code_t _gpgrt_nvc_parse (gpgrt_nvc_t *result, int *errlinep,
+                                 estream_t stream, unsigned int flags);
+gpg_err_code_t _gpgrt_nvc_write (gpgrt_nvc_t cont, estream_t stream);
+gpgrt_nve_t _gpgrt_nve_next (gpgrt_nve_t entry, const char *name);
+const char *_gpgrt_nve_name (gpgrt_nve_t entry);
+const char *_gpgrt_nve_value (gpgrt_nve_t entry);
+/* Convenience functions.  */
+const char *_gpgrt_nvc_get_string (gpgrt_nvc_t nvc, const char *name);
+int _gpgrt_nvc_get_bool (gpgrt_nvc_t nvc, const char *name);
+
+
+
 
 /*
- * Internal platform abstraction functions (sysutils.c)
+ * Internal platform abstraction functions (sysutils.c and stringutil.c)
  */
 
 /* Return true if FD is valid.  */
@@ -794,6 +830,9 @@ gpg_err_code_t _gpgrt_chdir (const char *name);
 /* Return the current WD as a malloced string.  */
 char *_gpgrt_getcwd (void);
 
+/* Wrapper for Windows to allow utf8 file names.  */
+gpg_err_code_t _gpgrt_access (const char *fname, int mode);
+
 /* Return the home directory of user NAME.  */
 char *_gpgrt_getpwdir (const char *name);
 
@@ -808,6 +847,9 @@ char *_gpgrt_fnameconcat (const char *first_part,
 char *_gpgrt_absfnameconcat (const char *first_part,
                              ... ) GPGRT_ATTR_SENTINEL(0);
 
+/* What the name says.  */
+char *_gpgrt_trim_spaces (char *str);
+
 
 /*
  * Platform specific functions (Windows)
@@ -817,7 +859,16 @@ char *_gpgrt_absfnameconcat (const char *first_part,
 char *_gpgrt_w32_reg_query_string (const char *root,
                                    const char *dir,
                                    const char *name);
+char *_gpgrt_w32_reg_get_string (const char *key);
 
+wchar_t *_gpgrt_fname_to_wchar (const char *fname);
+
+enum gpgrt_windows_feature
+  {
+    GPGRT_WINDOWS_PROC_ATTRIBUTE,
+    GPGRT_WINDOWS_UNDER_WINE
+  };
+int _gpgrt_windows_feature (int);
 
 #endif /*HAVE_W32_SYSTEM*/
 
